;;; bench3.el --- benchmark read-extended-command -*- lexical-binding: t -*-

;; Define lots of commands with a `completion-predicate'.

(defmacro bench-define-cmd (i)
  `(defun ,(intern (format "bench-cmd-%d" i)) ()
     (declare (completion always))
     (interactive)))

(defmacro bench-define-cmds (n)
  (let (cmds)
    (dotimes (i n)
      (push `(bench-define-cmd ,i) cmds))
    `(progn ,@cmds)))

(bench-define-cmds 16384)

;; Extended `read-extended-command' subroutines with safety nets.

(defun bench-completion-predicate (symbol buffer)
  (if (get symbol 'completion-predicate)
      (condition-case-unless-debug err
          (funcall (get symbol 'completion-predicate) symbol buffer)
        (error (message "bench-completion-predicate: %s: %S" symbol err)))
    (or (null (command-modes symbol))
        (command-completion-using-modes-p symbol buffer))))

(defun bench-read-extended-command ()
  (let ((buffer (current-buffer)))
    (minibuffer-with-setup-hook
        (lambda ()
          (add-hook 'post-self-insert-hook
                    (lambda ()
                      (setq execute-extended-command--last-typed
                            (minibuffer-contents)))
                    nil 'local)
          (setq-local minibuffer-default-add-function
	              (lambda ()
	                ;; Get a command name at point in the original buffer
	                ;; to propose it after M-n.
	                (let ((def
                               (with-current-buffer
			           (window-buffer (minibuffer-selected-window))
			         (and (commandp (function-called-at-point))
				      (format
                                       "%S" (function-called-at-point)))))
		              (all (sort (minibuffer-default-add-completions)
                                         #'string<)))
		          (if def
		              (cons def (delete def all))
		            all)))))
      ;; Read a string, completing from and restricting to the set of
      ;; all defined commands.  Don't provide any initial input.
      ;; Save the command read on the extended-command history list.
      (completing-read
       (concat (cond
	        ((eq current-prefix-arg '-) "- ")
	        ((and (consp current-prefix-arg)
		      (eq (car current-prefix-arg) 4)) "C-u ")
	        ((and (consp current-prefix-arg)
		      (integerp (car current-prefix-arg)))
	         (format "%d " (car current-prefix-arg)))
	        ((integerp current-prefix-arg)
	         (format "%d " current-prefix-arg)))
	       ;; This isn't strictly correct if `execute-extended-command'
	       ;; is bound to anything else (e.g. [menu]).
	       ;; It could use (key-description (this-single-command-keys)),
	       ;; but actually a prompt other than "M-x" would be confusing,
	       ;; because "M-x" is a well-known prompt to read a command
	       ;; and it serves as a shorthand for "Extended command: ".
               (if (memq 'shift (event-modifiers last-command-event))
	           "M-X "
	         "M-x "))
       (lambda (string pred action)
         (if (and suggest-key-bindings (eq action 'metadata))
	     '(metadata
	       (affixation-function . read-extended-command--affixation)
	       (category . command))
           (complete-with-action action obarray string pred)))
       (lambda (sym)
         (and (commandp sym)
              (or (null read-extended-command-predicate)
                  (and (functionp read-extended-command-predicate)
                       (condition-case-unless-debug err
                           (funcall read-extended-command-predicate sym buffer)
                         (error
                          (message "read-extended-command-predicate: %s: %S"
                                   sym err)))))))
       t nil 'extended-command-history))))

;; Benchmarks.

(defun bench-vanilla-no-pred ()
  (let ((read-extended-command-predicate nil))
    (read-extended-command)))

(defun bench-vanilla-with-pred ()
  (let ((read-extended-command-predicate
         #'command-completion-default-include-p))
    (read-extended-command)))

(defun bench-vanilla-safe-pred ()
  (let ((read-extended-command-predicate #'bench-completion-predicate))
    (read-extended-command)))

(defun bench-safe-no-pred ()
  (let ((read-extended-command-predicate nil))
    (bench-read-extended-command)))

(defun bench-safe-with-pred ()
  (let ((read-extended-command-predicate
         #'command-completion-default-include-p))
    (bench-read-extended-command)))

(with-current-buffer-window "*bench*" nil nil
  (let* ((loops 3)
         (fns '(bench-vanilla-no-pred
                bench-vanilla-with-pred
                bench-vanilla-safe-pred
                bench-safe-no-pred
                bench-safe-with-pred))
         (progress (make-progress-reporter
                    "Marking benches" 0 (* 2 loops loops (length fns)))))
    (pcase-dolist (`(,reps . ,keys)
                   '((10 . "\C-i\C-m")
                     (100 . "bench-\C-i\C-a\C-k\C-m")))
      (insert (if (bobp) "" ?\n) "== " (key-description keys) " ==\n")
      (setq keys (listify-key-sequence keys))
      (dotimes (_ loops)
        (insert ?\n)
        (dolist (fn fns)
          (dotimes (_ loops)
            (garbage-collect)
            (insert (apply #'format "%-24s %f %2d %f\n" fn
                           (benchmark-run reps
                             (setq unread-command-events keys)
                             (funcall fn))))
            (progress-reporter-update progress (car progress))))))
    (progress-reporter-done progress)))

;;; bench3.el ends here
